@@ -26,8 +26,7 @@
 
 use crate::{response, AccessControl};
 use futures_channel::mpsc;
-use futures_util::future::join_all;
-use futures_util::stream::StreamExt;
+use futures_util::{future::join_all, stream::StreamExt, FutureExt};
 use hyper::{
 	server::{conn::AddrIncoming, Builder as HyperBuilder},
 	service::{make_service_fn, service_fn},
@@ -49,7 +48,10 @@ use serde_json::value::RawValue;
 use socket2::{Domain, Socket, Type};
 use std::{
 	cmp,
-	net::{SocketAddr, TcpListener},
+	future::Future,
+	net::{SocketAddr, TcpListener, ToSocketAddrs},
+	pin::Pin,
+	task::{Context, Poll},
 };
 
 /// Builder to create JSON-RPC HTTP server.
@@ -104,7 +106,50 @@ impl Builder {
 	}
 
 	/// Finalizes the configuration of the server.
-	pub fn build(self, addr: SocketAddr) -> Result<Server, Error> {
+	///
+	/// ```rust
+	/// #[tokio::main]
+	/// async fn main() {
+	///   let listener = std::net::TcpListener::bind("127.0.0.1:0").unwrap();
+	///   let occupied_addr = listener.local_addr().unwrap();
+	///   let addrs: &[std::net::SocketAddr] = &[
+	///       occupied_addr,
+	///       "127.0.0.1:0".parse().unwrap(),
+	///   ];
+	///   assert!(jsonrpsee_http_server::HttpServerBuilder::default().build(occupied_addr).is_err());
+	///   assert!(jsonrpsee_http_server::HttpServerBuilder::default().build(addrs).is_ok());
+	/// }
+	/// ```
+	pub fn build(self, addrs: impl ToSocketAddrs) -> Result<Server, Error> {
+		let mut err: Option<Error> = None;
+
+		for addr in addrs.to_socket_addrs()? {
+			let (listener, local_addr) = match self.inner_builder(addr) {
+				Ok(res) => res,
+				Err(e) => {
+					err = Some(e);
+					continue;
+				}
+			};
+
+			return Ok(Server {
+				listener,
+				local_addr,
+				access_control: self.access_control,
+				max_request_body_size: self.max_request_body_size,
+				resources: self.resources,
+				tokio_runtime: self.tokio_runtime,
+			});
+		}
+
+		let err = err.unwrap_or_else(|| std::io::Error::new(std::io::ErrorKind::NotFound, "No address found").into());
+		Err(err)
+	}
+
+	fn inner_builder(
+		&self,
+		addr: SocketAddr,
+	) -> Result<(hyper::server::Builder<hyper::server::conn::AddrIncoming>, Option<SocketAddr>), Error> {
 		let domain = Domain::for_address(addr);
 		let socket = Socket::new(domain, Type::STREAM, None)?;
 		socket.set_nodelay(true)?;
@@ -117,17 +162,8 @@ impl Builder {
 		socket.listen(128)?;
 		let listener: TcpListener = socket.into();
 		let local_addr = listener.local_addr().ok();
-
 		let listener = hyper::Server::from_tcp(listener)?;
-
-		Ok(Server {
-			listener,
-			local_addr,
-			access_control: self.access_control,
-			max_request_body_size: self.max_request_body_size,
-			resources: self.resources,
-			tokio_runtime: self.tokio_runtime,
-		})
+		Ok((listener, local_addr))
 	}
 }
 
@@ -143,17 +179,17 @@ impl Default for Builder {
 	}
 }
 
-/// Handle used to stop the running server.
+/// Handle used to run or stop the server.
 #[derive(Debug)]
-pub struct StopHandle {
+pub struct ServerHandle {
 	stop_sender: mpsc::Sender<()>,
-	stop_handle: Option<tokio::task::JoinHandle<()>>,
+	pub(crate) handle: Option<tokio::task::JoinHandle<()>>,
 }
 
-impl StopHandle {
+impl ServerHandle {
 	/// Requests server to stop. Returns an error if server was already stopped.
 	pub fn stop(mut self) -> Result<tokio::task::JoinHandle<()>, Error> {
-		let stop = self.stop_sender.try_send(()).map(|_| self.stop_handle.take());
+		let stop = self.stop_sender.try_send(()).map(|_| self.handle.take());
 		match stop {
 			Ok(Some(handle)) => Ok(handle),
 			_ => Err(Error::AlreadyStopped),
@@ -161,6 +197,19 @@ impl StopHandle {
 	}
 }
 
+impl Future for ServerHandle {
+	type Output = ();
+
+	fn poll(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
+		let handle = match &mut self.handle {
+			Some(handle) => handle,
+			None => return Poll::Ready(()),
+		};
+
+		handle.poll_unpin(cx).map(|_| ())
+	}
+}
+
 /// An HTTP JSON RPC server.
 #[derive(Debug)]
 pub struct Server {
@@ -185,7 +234,7 @@ impl Server {
 	}
 
 	/// Start the server.
-	pub fn start(mut self, methods: impl Into<Methods>) -> Result<StopHandle, Error> {
+	pub fn start(mut self, methods: impl Into<Methods>) -> Result<ServerHandle, Error> {
 		let max_request_body_size = self.max_request_body_size;
 		let access_control = self.access_control;
 		let (tx, mut rx) = mpsc::channel(1);
@@ -222,7 +271,7 @@ impl Server {
 							Err(GenericTransportError::TooLarge) => return Ok::<_, HyperError>(response::too_large()),
 							Err(GenericTransportError::Malformed) => return Ok::<_, HyperError>(response::malformed()),
 							Err(GenericTransportError::Inner(e)) => {
-								log::error!("Internal error reading request body: {}", e);
+								tracing::error!("Internal error reading request body: {}", e);
 								return Ok::<_, HyperError>(response::internal_error());
 							}
 						};
@@ -236,7 +285,9 @@ impl Server {
 						if is_single {
 							if let Ok(req) = serde_json::from_slice::<Request>(&body) {
 								// NOTE: we don't need to track connection id on HTTP, so using hardcoded 0 here.
-								if let Some(fut) = methods.execute_with_resources(&tx, req, 0, &resources) {
+								if let Some(fut) =
+									methods.execute_with_resources(&tx, req, 0, &resources, max_request_body_size)
+								{
 									fut.await;
 								}
 							} else if let Ok(_req) = serde_json::from_slice::<Notif>(&body) {
@@ -249,11 +300,9 @@ impl Server {
 						// Batch of requests or notifications
 						} else if let Ok(batch) = serde_json::from_slice::<Vec<Request>>(&body) {
 							if !batch.is_empty() {
-								join_all(
-									batch
-										.into_iter()
-										.filter_map(|req| methods.execute_with_resources(&tx, req, 0, &resources)),
-								)
+								join_all(batch.into_iter().filter_map(|req| {
+									methods.execute_with_resources(&tx, req, 0, &resources, max_request_body_size)
+								}))
 								.await;
 							} else {
 								// "If the batch rpc call itself fails to be recognized as an valid JSON or as an
@@ -281,7 +330,7 @@ impl Server {
 						} else {
 							collect_batch_response(rx).await
 						};
-						log::debug!("[service_fn] sending back: {:?}", &response[..cmp::min(response.len(), 1024)]);
+						tracing::debug!("[service_fn] sending back: {:?}", &response[..cmp::min(response.len(), 1024)]);
 						Ok::<_, HyperError>(response::ok_response(response))
 					}
 				}))
@@ -298,7 +347,7 @@ impl Server {
 			let _ = server.with_graceful_shutdown(async move { rx.next().await.map_or((), |_| ()) }).await;
 		});
 
-		Ok(StopHandle { stop_handle: Some(handle), stop_sender: tx })
+		Ok(ServerHandle { handle: Some(handle), stop_sender: tx })
 	}
 }
 
